/*	pbpbit.h

	Pattern Bit layer for PBP; includes quantum-style operators.

	See pbp.h for copyright, etc.
*/

class pbit
{
	REreg_t v;  // value

public:
	// Constructor
	pbit(int value=pbitNaN) {
		v = (((value < 0) || (value >= pbitNaN)) ? pbitNaN : value);
	}

	// Helpers
	void Show() const {
		if (v < pbitNaN) re.Show(v);
		else fprintf(stderr, "pbitNaN\n");
	}

	int Valid() const {
		return((v >= 0) && (v < pbitNaN));
	}

	// Operators
	void operator=(const pbit& a) {
		v = (a.Valid() ? a.v : pbitNaN);
	}

	pbit And(const pbit& a) const {
		return((a.Valid() && Valid()) ? pbit(re.And(a.v, v)) : pbit(pbitNaN));
	}
	pbit operator&(const pbit& a) const { return(And(a)); }
	pbit operator*(const pbit& a) const { return(And(a)); }

	pbit Or(const pbit& a) const {
		return((a.Valid() && Valid()) ? pbit(re.Or(a.v, v)) : pbit(pbitNaN));
	}
	pbit operator|(const pbit& a) const { return(Or(a)); }

	pbit Xor(const pbit& a) const {
		return((a.Valid() && Valid()) ? pbit(re.Xor(a.v, v)) : pbit(pbitNaN));
	}
	pbit operator^(const pbit& a) const { return(Xor(a)); }
	pbit operator+(const pbit& a) const { return(Xor(a)); }

	// 1-bit ~ is same as !
	pbit Not() const { return(Valid() ? pbit(re.Xor(1, v)) : pbit(pbitNaN)); }
	pbit operator~() const { return(Not()); }
	pbit operator!() const { return(Not()); }

	pbit Sub(const pbit& a) const {
		// this - a means this & ~a
		return((a.Valid() && Valid()) ? pbit(re.And(v, re.Xor(1, a.v))) : pbit(pbitNaN));
	}
	pbit operator-(const pbit& a) const { return(Sub(a)); }

	pbit LT(const pbit& a) const {
		// this < a means a & ~this
		return((a.Valid() && Valid()) ? pbit(re.And(a.v, re.Xor(1, v))) : pbit(pbitNaN));
	}
	pbit operator<(const pbit& a) const { return(LT(a)); }

	pbit GT(const pbit& a) const { return(a < pbit(v)); }
	pbit operator>(const pbit& a) const { return(GT(a)); }

	pbit LE(const pbit& a) const {
		// this <= a means ~(this > a)
		return((a.Valid() && Valid()) ? pbit(re.Xor(1, GT(a).v)) : pbit(pbitNaN));
	}
	pbit operator<=(const pbit& a) const { return(LE(a)); }

	pbit GE(const pbit& a) const { return(a <= pbit(v)); }
	pbit operator>=(const pbit& a) const { return(GE(a)); }

	pbit NE(const pbit& a) const { return(Xor(a)); }
	pbit operator!=(const pbit& a) const { return(NE(a)); }

	pbit EQ(const pbit& a) const {
		// this == a means ~(this != a)
		return((a.Valid() && Valid()) ? pbit(re.Xor(1, NE(a).v)) : pbit(pbitNaN));
	}
	pbit operator==(const pbit& a) const { return(EQ(a)); }

	pbit Rot(const int a) const {
		return(Valid() ? pbit(re.Rot(v, a)) : pbit(pbitNaN));
	}
	pbit operator<<(const int a) const { return(Rot(a)); }

	pbit Flip(const int a) const {
		return(Valid() ? pbit(re.Flip(v, a)) : pbit(pbitNaN));
	}

	pbit Reset(const int a) const { return(Valid() ? pbit(re.Reset(v, a)) : pbit(pbitNaN)); }

	pbit Set(const int a) const { return(Valid() ? pbit(re.Set(v, a)) : pbit(pbitNaN)); }

	pbit Tog(const int a) const { return(Valid() ? pbit(re.Tog(v, a)) : pbit(pbitNaN)); }

	pbit Dom(const int a) const { return(Valid() ? pbit(re.Dom(v, a)) : pbit(pbitNaN)); }

	pbit Meas(const int a) const { return(Valid() ? pbit(re.Meas(v, a)) : pbit(pbitNaN)); }

	pbit Any() const { return(Valid() ? pbit(re.Any(v)) : pbit(pbitNaN)); }

	pbit All() const { return(Valid() ? pbit(re.All(v)) : pbit(pbitNaN)); }

	REreg_t First() const { return(Valid() ? re.First(v) : pbitNaN); }

	REreg_t Ones() const { return(Valid() ? re.Ones(v) : pbitNaN); }

	pbit Select(const pbit& a, const pbit& b) const {
		// a few simplifications
		if (v == 0) return(b);
		if (v == 1) return(a);
		if (a.v == b.v) return(a);
		if ((!Valid()) || (!a.Valid()) || (!b.Valid())) return(pbitNaN);

		// do the bit ops
		return((a & *this) | (b & ~*this));
	}

	friend class pint;
	friend class pfloat;

	// quantum-style friends...
	friend void NOT(pbit& q);
	friend void CNOT(const pbit& c, pbit& t);
	friend void SWAP(pbit& i0, pbit& i1);
	friend void CCNOT(const pbit& a, const pbit& b, pbit& c);
	friend void CSWAP(const pbit& c, pbit& i0, pbit& i1);
	friend void H(pbit& q, const int c);
	friend void SETMEAS(const int m);
	friend void SETMEAS();
	friend int MEAS(pbit& q);
};

// quantum-style pbit operators
// all these operate on pbits in-place
void NOT(pbit& q) { q = pbit(1) ^ q; }
void CNOT(const pbit& c, pbit& t) { t = c ^ t; }
void SWAP(pbit& i0, pbit& i1) { pbit t = i0; i0 = i1; i1 = t; }
void CCNOT(const pbit& a, const pbit& b, pbit& c) { c = c ^ (a & b); }
void CSWAP(const pbit& c, pbit& i0, pbit& i1) { pbit t = c.Select(i1,i0); i1 = c.Select(i0,i1); i0 = t; }
void H(pbit& q, const int c) { q = q ^ pbit(c + 2); }
int MEASat = 0;
void SETMEAS(const int m) { MEASat = (m & (REBITS - 1)); }
void SETMEAS() { SETMEAS(rand()); }
int MEAS(pbit& q) { q = q.Meas(MEASat); return(q.v); }
