/*	pbppfloat.h

	Pattern FLOATing point layer for PBP.

	This is not your usual float in that it is variable
	precision in both exponent and fractional part. The
	maximum precision for the fractional part is set by
	the user, but exponent dynamically expands/contracts.
	This version uses conventional normalization to the
	top bit of the fraction.

	See pbp.h for copyright, etc.
*/

int
sign_of(const float f)
{
	// return sign bit from a float
	return(f < 0);
}

int
exp_of(const float f)
{
	// return exponent of a float
	union { float f; int i; } fi;
	fi.f = f;
	return(((fi.i >> 23) & 0xff) - 128);
}

int
frac_of(const float f)
{
	// return fraction of a float (with hidden MSB 1)
	if (f == 0) return(0);
	union { float f; int i; } fi;
	fi.f = f;
	return((fi.i & ((1 << 23) - 1)) + (1 << 23));
}

#define	FRACBITS 3

class pfloat
{
	pint	sign;	// sign bit
	pint	exp;	// exponent
	pint	frac;	// fractional part
	uint8_t	fbits;	// fractional part bits limit

public:
	// constructors
	pfloat(const int fracbits = FRACBITS) {
		// defaults to bfloat16-ish 0
		sign = pint(0);
		exp = pint(0);
		frac = pint(0);
		fbits = fracbits;
	}

	pfloat(const float f, const int fracbits = FRACBITS) {
		// defaults to bfloat16-ish
		sign = pint(sign_of(f));
		exp = pint(exp_of(f));
		frac = pint((frac_of(f) >> (24 - fracbits)), fracbits);
		fbits = fracbits;
		*this = Normalize();
	}

	pfloat(const double d, const int fracbits = FRACBITS) {
		// defaults to bfloat16-ish
		float f = d; // enough with the doubles
		sign = pint(sign_of(f));
		exp = pint(exp_of(f));
		frac = pint((frac_of(f) >> (24 - fracbits)), fracbits);
		fbits = fracbits;
		*this = Normalize();
	}

	// Logic
	pint LAnd(const pfloat& b) const;
	pint LOr(const pfloat& b) const;
	pint LXOr(const pfloat& b) const;

	// Equality
	pint EQ(const pfloat& b) const;
	pint NE(const pfloat& b) const;

	// Relational
	pint GT(const pfloat& b) const;
	pint LT(const pfloat& b) const;
	pint GE(const pfloat& b) const;
	pint LE(const pfloat& b) const;

	// Shifts
	pfloat ShR(const pint& b) const;
	pfloat ShL(const pint& b) const;

	// Additive
	pfloat Add(pfloat b) const;
	pfloat Sub(pfloat b) const;

	// Multiplicative
	pfloat Mul(const pfloat& b) const;
	pfloat Recip(int iter) const;
	pfloat Div(const pfloat& b, int iter) const;

	// Not
	pint LNot() const;
	pint Logic() const;

	// Sign manipulation
	pfloat Neg() const;
	pfloat Abs() const;

	// PBP primitives
	pfloat Rot(const int b) const;
	pfloat Flip(const int b) const;
	pfloat Reset(const int b) const;
	pfloat Set(const int e, const float v) const;
	pfloat Dom(const int b) const;
	float Meas(const int b) const;
	float Meas() const;
	REreg_t First() const;
	REreg_t Pop() const;

	// Helpers
//	int Valid() const;
	pfloat Normalize() const;
	pfloat Select(const pfloat& vtrue, const pfloat& vfalse) const;

	// Extended functions
	pfloat Min(const pfloat& b) const;
	pfloat Max(const pfloat& b) const;
	pfloat Exp() const;
	pfloat Cos() const;
	pfloat Sin() const;
	pfloat Tan() const;
	pfloat ArcTan() const;
	pfloat Log() const;
	pfloat Sqrt() const;

	// Reductions
	int Any() const;
	int All() const;
	float ReduceAdd() const;
	float ReduceMul() const;
	float ReduceMin() const;
	float ReduceMax() const;

	// Scans (parallel prefix)
	pfloat ScanAdd() const;
	pfloat ScanMul() const;
	pfloat ScanMin() const;
	pfloat ScanMax() const;

	// Gather
	int Gather(float *p, const int n) const;

	// Display
	void Show(const int vals) const;

	// Friends
	friend class pint;
	friend pfloat Range(const float min, const float max, const int fracr,
			    const int dim, const int fracbits);
	friend pfloat Scatter(const float *p, const int n, const int fracbits);

	// Expensive library stuff
	pfloat SortUp() const;
	pfloat SortDown() const;

	// Assignment operators
	pfloat operator=(const pfloat& b) {
		sign = b.sign;
		exp = b.exp;
		frac = b.frac;
		fbits = b.fbits;
		return(*this);
	}
	pfloat operator*=(pfloat b) { *this = Mul(b); return(*this); }
	pfloat operator/=(pfloat b) { *this = Div(b,-1); return(*this); }
	pfloat operator+=(pfloat b) { *this = Add(b); return(*this); }
	pfloat operator-=(pfloat b) { *this = Sub(b); return(*this); }
	pfloat operator>>=(pint b) { *this = ShR(b); return(*this); }
	pfloat operator<<=(pint b) { *this = ShL(b); return(*this); }

	// Logical operators
	pint operator&&(const pfloat& b) const { return(LAnd(b)); }
	pint operator||(const pfloat& b) const { return(LOr(b)); }
	// there is no operator^^, but != effectively does it

	// Equality operators
	pint operator==(const pfloat& b) const { return(EQ(b)); }
	pint operator!=(const pfloat& b) const { return(NE(b)); }

	// Relational operators
	pint operator>(const pfloat& b) const { return(GT(b)); }
	pint operator<(const pfloat& b) const { return(LT(b)); }
	pint operator>=(const pfloat& b) const { return(GE(b)); }
	pint operator<=(const pfloat& b) const { return(LE(b)); }

	// C++ 20 three-way comparison operator
	// uncomment if your C++ supports it (GCC doesn't yet)
//	pint operator<=>(const pfloat& b) const { return(GT(b) - LT(b)); }

	// Bitwise shift operators (divide/multiply by power of 2)
	pfloat operator>>(pint b) const { return(ShR(b)); }
	pfloat operator<<(pint b) const { return(ShL(b)); }

	// Additive operators
	pfloat operator+(pfloat b) const { return(Add(b)); }
	pfloat operator-(pfloat b) const { return(Sub(b)); }

	// Multiplicative operators
	pfloat operator*(pfloat b) const { return(Mul(b)); }
	pfloat operator/(pfloat b) const { return(Div(b,-1)); }

	// Not operators
	pint operator!() const { return(LNot()); }

	// Unary plus and minus
	pfloat operator+() const { return(*this); }
	pfloat operator-() const { return(Neg()); }

	// Prefix increment and decrement
	pfloat& operator++() { *this = Add(pfloat((float) 1.0)); return(*this); }
	pfloat& operator--() { *this = Sub(pfloat((float) 1.0)); return(*this); }

	// Postfix increment and decrement
	// note that int arg is a C++ notation thing; there isn't an int arg
	pfloat operator++(int) { pfloat r = *this; *this = Add(pfloat((float) 1.0)); return(r); }
	pfloat operator--(int) { pfloat r = *this; *this = Sub(pfloat((float) 1.0)); return(r); }
};

inline
pint::operator pfloat() const {
	// type cast pint to pfloat
	pfloat f;
	f.sign = (((*this).type == UPINT) ? pint(0) : (*this).Bit(prec-1));
	f.exp = pint(6);
	f.frac = f.sign.Select(-*this, *this);
	return(f.Normalize());
}

inline pfloat
Range(const float min, const float max, const int fracr, const int dim = 0, const int fracbits = FRACBITS)
{
	// create a pfloat representing a range covering [min, max]
	// min and max are treated as positive
	// fracr is the number of bits to vary (top bits)
	// mask selects the entanglement channels to use
	// fracbits is (optional) bits precision of the pfloat

	// create the entangled bits
	pint hbits = pint(0).Had(fracr, dim);

	// convert to float
	pfloat r = hbits;

	// scale it
	int imax = (1 << fracr) - 1;
	float range = (max - min);
	r = r * pfloat(range / imax);
	return(r + pfloat(min));
}

pfloat
Scatter(const float *p, const int n, const int fracbits = FRACBITS)
{
	// Scatter a host array into the
	// entanglement channels of a pfloat
	pfloat r;
	int *t;
	t = new int[n];
	for (int i=0; i<n; ++i) t[i] = sign_of(p[i]);
	r.sign = Scatter(&(t[0]), n, 1);
	for (int i=0; i<n; ++i) t[i] = exp_of(p[i]);
	r.exp = Scatter(&(t[0]), n);
	for (int i=0; i<n; ++i) t[i] = (frac_of(p[i]) >> (24 - fracbits));
	r.frac = Scatter(&(t[0]), n, fracbits);
	r.fbits = fracbits;
	delete[] t;
	return(r.Normalize());
}

// Logic

pint
pfloat::LAnd(const pfloat& b) const
{
	// Logical And only needs compare fracs
	return(frac.LAnd(b.frac));
}

pint
pfloat::LOr(const pfloat& b) const
{
	// Logical Or only needs compare fracs
	return(frac.LOr(b.frac));
}

pint
pfloat::LXOr(const pfloat& b) const
{
	// Logical XOr only needs compare fracs
	return(frac.LXOr(b.frac));
}

pint
pfloat::EQ(const pfloat& b) const
{
	pint r = ((sign != b.sign) |
		  (exp != b.exp) |
		  (frac != b.frac));
	r.bit[0] = ~r.bit[0];
	return(r);
}

pint
pfloat::NE(const pfloat& b) const
{
	pint r = ((sign != b.sign) |
		  (exp != b.exp) |
		  (frac != b.frac));
	return(r);
}

// Relational

pint
pfloat::GT(const pfloat& b) const
{
	// float GT is really first NE is GT in sign, exp, or frac
	pint gt = (b.sign & !sign); // positive > negative
	pint eq = (sign == b.sign); // same sign
	gt = eq.Select((exp > b.exp), gt);
	eq = eq & (exp == b.exp); // same sign & exp
	gt = eq.Select((frac > b.frac), gt);
	return(gt);
}

pint
pfloat::LT(const pfloat& b) const
{
	return(b.GT(*this));
}

pint
pfloat::GE(const pfloat& b) const
{
	return(!LT(b));
}

pint
pfloat::LE(const pfloat& b) const
{
	return(!GT(b));
}

// Shifts

pfloat
pfloat::ShR(const pint& b) const
{
	pfloat r = *this;
	r.exp = r.exp - b;
	return(r);
}

pfloat
pfloat::ShL(const pint& b) const
{
	pfloat r = *this;
	r.exp = r.exp + b;
	return(r);
}

// Additive

pfloat
pfloat::Add(pfloat b) const
{
	pfloat a = *this;

	// fix a's exponent if it is smaller
	pint tonorm = (a.frac != 0) & (a.exp < b.exp);
	if (tonorm.Any()) {
		// shift whole distance in one step
		pint edif = tonorm.Select((b.exp - a.exp), pint(0));
		a.exp = a.exp + edif;
		a.frac = (a.frac >> edif);
		// make 0 frac be truly 0, but with b's exponent
		tonorm = (a.frac != pint(0));
		a.exp = tonorm.Select(a.exp, b.exp);
		a.sign = tonorm * a.sign;
	}

	// fix b's exponent if it is smaller
	tonorm = (b.frac != 0) & (b.exp < a.exp);
	if (tonorm.Any()) {
		// shift whole distance in one step
		pint edif = tonorm * (a.exp - b.exp);
		b.exp = b.exp + edif;
		b.frac = (b.frac >> edif);
		// make 0 frac be truly 0, but with a's exponent
		tonorm = (b.frac != pint(0));
		b.exp = tonorm.Select(b.exp, a.exp);
		b.sign = tonorm * b.sign;
	}

	// add them
	if ((a.sign != b.sign).Any()) {
		// make them signed, then add
		a.frac = a.sign.Select((pint(0) - a.frac), a.frac);
		b.frac = b.sign.Select((pint(0) - b.frac), b.frac);
		a.frac = a.frac + b.frac;
		a.sign = (a.frac < pint(0));
		a.frac = a.sign.Select((pint(0) - a.frac), a.frac);
	} else {
		// easy; both the same sign
		a.frac = a.frac + b.frac;
	}

	// renormalize result
	return(a.Normalize());
}

pfloat
pfloat::Sub(pfloat b) const
{
	b = Add(b.Neg());
	return(b);
}

// Multiplicative

pfloat
pfloat::Mul(const pfloat& b) const
{
	pfloat r = *this;
	r.sign ^= b.sign;
	r.exp += b.exp;
	r.frac = (frac.Mul(b.frac) >> (r.fbits-2));
	return(r.Normalize());
}

pfloat
pfloat::Recip(int iter = -1) const
{
	// Reciprocal using Newton-Raphson iterations
	// iter >= 0 give number of iterations
	// iter < 0 automatically picks based on fbits
	pfloat r = *this;

	// initial guess is negate exponent
	r.exp = pint(-2) - r.exp; // exp is correct
	r.frac = pint(1 << (r.fbits-1)); // make mantissa 1
	r = r.Normalize();

	// do the iterations
	if (iter < 0) iter = fbits / 2;
	for (int i=0; i<iter; ++i) {
		r = r * (pfloat((float) 2.0) - (r * *this));
	}
	return(r);
}

pfloat
pfloat::Div(const pfloat& b, int iter = -1) const
{
	// Divide is really Mul by Recip
	return(Mul(b.Recip(iter)));
}

// Not

pint
pfloat::LNot() const
{
	// true iff frac==0
	return(frac.Not());
}

pint
pfloat::Logic() const
{
	// true iff frac!=0
	return(frac.Logic());
}

// Sign manipulation

pfloat
pfloat::Neg() const
{
	pfloat r = *this;
	r.sign = !r.sign;
	return(r);
}

pfloat
pfloat::Abs() const
{
	pfloat r = *this;
	r.sign.bit[0] = pbit(0);
	return(r);
}

// PBP primitives

pfloat
pfloat::Rot(const int b) const
{
	pfloat r = *this;
	r.sign = sign.Rot(b);
	r.exp = exp.Rot(b);
	r.frac = frac.Rot(b);
	return(r);
}

pfloat
pfloat::Flip(const int b) const
{
	pfloat r = *this;
	r.sign = sign.Flip(b);
	r.exp = exp.Flip(b);
	r.frac = frac.Flip(b);
	return(r);
}

pfloat
pfloat::Dom(const int b) const
{
	pint bit = frac.Logic();
	bit = bit.Dom(b);
	pfloat r;
	r = bit;
	return(r);
}

pfloat
pfloat::Reset(const int e) const
{
	return(Set(e, 0.0));
}

pfloat
pfloat::Set(const int e, const float v = 1.0) const
{
	pfloat r = *this;
	r.sign = r.sign.Set(e, sign_of(v));
	r.exp = r.exp.Set(e, exp_of(v));
	r.frac = r.frac.Set(e, frac_of(v));
	return(r.Normalize());
}

float
pfloat::Meas(const int b) const
{
	// get float value of entanglement channel b
	float r = frac.Meas(b) * powf(2.0, exp.Meas(b)-(frac.prec-2));
	if (sign.Meas(b)) r = -r;
	return(r);
}

float
pfloat::Meas() const
{
	// Quantum-like random measurement
	return(Meas(rand() & (REBITS - 1)));
}

REreg_t
pfloat::First() const
{
	// first is first non-0 value...
	// which is just first non-0 frac
	return(frac.First());
}

REreg_t
pfloat::Pop() const
{
	// pop(ulation) is number of non-zero values
	// which is just Pop of frac
	return(frac.Pop());
}

// Helpers

pfloat
pfloat::Normalize() const
{
	// normalize
	pfloat r = *this;
	r.frac = r.frac.Minimize(); // clip leading 0s
	pint tonorm(1);

	// shift down any with too many bits
	while (r.frac.prec > fbits) {
		// make pint out of top fraction bit
		tonorm.bit[0] = r.frac.bit[r.frac.prec-1];
		// bump exponents
		r.exp = r.exp + tonorm;
		// shift them down
		r.frac = (r.frac >> tonorm);
	}

	// make 0 fracs be truly 0
	tonorm = (r.frac != pint(0));
	r.exp = tonorm.Select(r.exp, pint(EXPOF0));
	r.sign = tonorm * r.sign;

	// shift up any non-0 value too small
	r.frac = r.frac.Extend(fbits);
	tonorm.bit[0] = tonorm.bit[0] & ~r.frac.bit[fbits-1];
	while (tonorm.Any()) {
		r.exp = r.exp - tonorm; // dec exponent on too small ones
		r.frac = (r.frac << tonorm).Extend(fbits);
		tonorm.bit[0] = tonorm.bit[0] & ~r.frac.bit[fbits-1];
	}
	return(r);
}

pfloat
pint::Select(const pfloat& vtrue, const pfloat& vfalse) const
{
	// multiplexor
	pfloat r = vtrue;
	r.sign = Select(vtrue.sign, vfalse.sign);
	r.exp = Select(vtrue.exp, vfalse.exp);
	r.frac = Select(vtrue.frac, vfalse.frac);
	return(r);
}

pfloat
pfloat::Select(const pfloat& vtrue, const pfloat& vfalse) const
{
	return(frac.Select(vtrue, vfalse));
}

// Extended functions
// transcendentals can follow formulas found on the WWW
// for example, http://www.netlib.org/fdlibm/
// but just straight Taylor or Maclaurin series terms for now
// e.g., https://people.math.sc.edu/girardi/m142/handouts/10sTaylorPolySeries.pdf
// https://en.wikipedia.org/wiki/Taylor_series

pfloat
pfloat::Min(const pfloat& b) const
{
	pfloat r = *this;
	pint gt = (r > b);
	r.sign = gt.Select(b.sign, r.sign);
	r.exp = gt.Select(b.exp, r.exp);
	r.frac = gt.Select(b.frac, r.frac);
	return(r);
}

pfloat
pfloat::Max(const pfloat& b) const
{
	pfloat r = *this;
	pint gt = (r > b);
	r.sign = gt.Select(r.sign, b.sign);
	r.exp = gt.Select(r.exp, b.exp);
	r.frac = gt.Select(r.frac, b.frac);
	return(r);
}

pfloat
pfloat::Exp() const
{
	// e to any pfloat power
	pfloat r(1.0);
	pfloat p = *this * *this;
	r = r + *this;
	r = r + (p * pfloat(1.0 / (2.0)));
	p = p * *this;
	r = r + (p * pfloat(1.0 / (3.0 * 2.0)));
	p = p * *this;
	r = r + (p * pfloat(1.0 / (4.0 * 3.0 * 2.0)));
	p = p * *this;
	r = r + (p * pfloat(1.0 / (5.0 * 4.0 * 3.0 * 2.0)));
	return(r);
}

pfloat
pfloat::Cos() const
{
	// cosine of any pfloat
	// alternatively could be (*this + pfloat(3.141592654/2.0)).Sin()
	pfloat r(1.0);
	pfloat s = (*this * *this);
	r = r - (s * pfloat(1.0 / (2.0)));
	pfloat pows = (s * s);
	r = r + (pows * pfloat(1.0 / (4.0*3.0*2.0)));
	pows = pows * s;
	r = r - (pows * pfloat(1.0 / (6.0*5.0*4.0*3.0*2.0)));
	pows = pows * s;
	r = r + (pows * pfloat(1.0 / (8.0*7.0*6.0*5.0*4.0*3.0*2.0)));
	return(r);
}

pfloat
pfloat::Sin() const
{
	// sine of any pfloat
	// alternatively could be (*this + pfloat(1.5*3.141592654)).Cos()
	pfloat r = *this;
	pfloat s = (r * r);
	pfloat pows = (r * s);
	r = r - (pows * pfloat(1.0 / (3.0*2.0)));
	pows = pows * s;
	r = r + (pows * pfloat(1.0 / (5.0*4.0*3.0*2.0)));
	pows = pows * s;
	r = r - (pows * pfloat(1.0 / (7.0*6.0*5.0*4.0*3.0*2.0)));
	pows = pows * s;
	r = r + (pows * pfloat(1.0 / (9.0*8.0*7.0*6.0*5.0*4.0*3.0*2.0)));
	return(r);
}

pfloat
pfloat::Tan() const
{
	// tangent of any pfloat in [-Pi/2, Pi/2]
	// terms listed at https://www.symbolab.com/solver/step-by-step/taylor%20tan%5Cleft(x%5Cright)
	// this gets way off...
	pfloat r = *this;
	pfloat s = (r * r);
	pfloat pows = (r * s);
	r = r - (pows * pfloat(1.0 / 3.0));
	pows = pows * s;
	r = r + (pows * pfloat(2.0 / 15.0));
	pows = pows * s;
	r = r - (pows * pfloat(17.0 / 315.0));
	pows = pows * s;
	r = r + (pows * pfloat(62.0 / 2835.0));
	return(r);
}

pfloat
pfloat::ArcTan() const
{
	// arctangent (inverse tangent) of any pfloat in [-1, 1]
	pfloat r = *this;
	pfloat s = (r * r);
	pfloat pows = (r * s);
	r = r - (pows * pfloat(1.0 / 3.0));
	pows = pows * s;
	r = r + (pows * pfloat(1.0 / 5.0));
	pows = pows * s;
	r = r - (pows * pfloat(1.0 / 7.0));
	pows = pows * s;
	r = r + (pows * pfloat(1.0 / 9.0));
	return(r);
}

pfloat
pfloat::Log() const
{
	// log base e (natural log, aka, ln) of any pfloat in (0, 2)
	// this is complicated by different formulas for values
	// less than or greater than 1, merged here because they
	// only differ in +/- of every other term
	pint lt = (*this < pfloat(1.0));
	pfloat gtx = (*this - pfloat(1.0));
	pfloat ltx = (pfloat(1.0) - *this);
	pfloat x;
	x.sign = lt.Select(ltx.sign, gtx.sign);
	x.exp = lt.Select(ltx.exp, gtx.exp);
	x.frac = lt.Select(ltx.frac, gtx.frac);
	pfloat r = x;
	r.sign = lt.Select(!x.sign, x.sign);
	pfloat s = x * x;
	r = r - (s * pfloat(1.0 / (2.0)));
	s = s * x;
	pfloat t = (s * pfloat(1.0 / (3.0 * 2.0)));
	t.sign = lt.Select(!t.sign, t.sign);
	r = r + t;
	s = s * x;
	r = r - (s * pfloat(1.0 / (4.0 * 3.0 * 2.0)));
	s = s * x;
	t = (s * pfloat(1.0 / (5.0 * 4.0 * 3.0 * 2.0)));
	t.sign = lt.Select(!t.sign, t.sign);
	r = r + t;
	s = s * x;
	r = r - (s * pfloat(1.0 / (6.0 * 5.0 * 4.0 * 3.0 * 2.0)));
	s = s * x;
	t = (s * pfloat(1.0 / (7.0 * 6.0 * 5.0 * 4.0 * 3.0 * 2.0)));
	t.sign = lt.Select(!t.sign, t.sign);
	r = r + t;
	return(r);
}

pfloat
pfloat::Sqrt() const
{
	// square root of any pfloat in (-1, 1)
	pfloat x = (*this - pfloat(1.0));
	pfloat r(1.0);
	pfloat p = x;
	r = r + (p * pfloat(1.0 / 2.0));
	p = p * x;
	r = r - (p * pfloat(1.0 / 8.0));
	p = p * x;
	r = r + (p * pfloat(1.0 / 16.0));
	p = p * x;
	r = r - (p * pfloat(5.0 / 128.0));
	p = p * x;
	r = r - (p * pfloat(7.0 / 256.0));
	return(r);
}

// Reductions

int
pfloat::Any() const
{
	return(frac.Any());
}

int
pfloat::All() const
{
	return(frac.All());
}

float
pfloat::ReduceAdd() const
{
	// Reduce Add by classical tree reduction
	// does large steps first because that makes
	// more of the chunks 0 early
	pfloat r = *this;
	for (int i=REWAYS-1; i>0; i>>=1) {
		pfloat toadd = r.Rot(1<<i);
		pint mask = pint(0).Dom(1<<i);
		r = r + mask.Select(toadd, pfloat(0.0));
	}
	return(r.Meas(0));
}

float
pfloat::ReduceMul() const
{
	// Reduce Mul by classical tree reduction
	// does large steps first because that makes
	// more of the chunks 0 early
	// Be careful! It's easy to have this blow up
	// and pfloat exponents can get huge
	pfloat r = *this;
	for (int i=REWAYS-1; i>0; i>>=1) {
		pfloat tomul = r.Rot(1<<i);
		pint mask = pint(0).Dom(1<<i);
		r = r * mask.Select(tomul, pfloat(0.0));
	}
	return(r.Meas(0));
}

float
pfloat::ReduceMin() const
{
	// Minimum reduction
	pfloat r = *this;
	int ansexp, ansfrac;
	if (r.sign.Any()) {
		// negative is less than positive
		r.exp = r.sign.Select(r.exp, pint(EXPOF0));
		r.frac = r.sign.Select(r.frac, pint(0));
		// zero frac where not largest exponent
		ansexp = r.exp.ReduceMax();
		r.frac = (r.exp == pint(ansexp)).Select(r.frac, pint(0));
		ansfrac = -r.frac.ReduceMax();
	} else {
		// look for smallest (they're all positive)
		// zero frac where not smallest exponent
		int ansexp = r.exp.ReduceMin();
		r.frac = (r.exp == pint(ansexp)).Select(r.frac, pint(0));
		int ansfrac = r.frac.ReduceMin();
	}
	return(ansfrac * powf(2.0, ansexp-(r.frac.prec-2)));
}

float
pfloat::ReduceMax() const
{
	// Minimum reduction
	pfloat r = *this;
	int ansexp, ansfrac;
	if (r.sign.All()) {
		// look for smalest (they're all negative)
		// zero frac where not largest exponent
		ansexp = r.exp.ReduceMin();
		r.frac = (r.exp == pint(ansexp)).Select(r.frac, pint(0));
		ansfrac = -r.frac.ReduceMax();
	} else {
		// positive is more than negative
		r.exp = r.sign.Select(pint(EXPOF0), r.exp);
		r.frac = r.sign.Select(pint(0), r.frac);

		// zero frac where not largest exponent
		int ansexp = r.exp.ReduceMax();
		r.frac = (r.exp == pint(ansexp)).Select(r.frac, pint(0));
		int ansfrac = r.frac.ReduceMax();
	}
	return(ansfrac * powf(2.0, ansexp-(r.frac.prec-2)));
}

// Scans (parallel prefix)

pfloat
pfloat::ScanAdd() const
{
	// unary scan Add (parallel prefix)
	pfloat r = *this;
	for (int i=0; i<REWAYS; ++i) {
		// send value up to i-dim neighbor
		pint mask = pint(1).Dom((1 << i) - 1);
		pfloat toadd = mask.Select(r.Rot(1 << i), pfloat(0.0));
		r = r + toadd;
	}
	return(r);
}

pfloat
pfloat::ScanMul() const
{
	// unary scan Mul (parallel prefix)
	pfloat r = *this;
	for (int i=0; i<REWAYS; ++i) {
		// send value up to i-dim neighbor
		pint mask = pint(1).Dom((1 << i) - 1);
		pfloat tomul = mask.Select(r.Rot(1 << i), pfloat(0.0));
		r = r * tomul;
	}
	return(r);
}

pfloat
pfloat::ScanMin() const
{
	// unary scan Min (parallel prefix)
	pfloat r = *this;
	for (int i=0; i<REWAYS; ++i) {
		// send value up to i-dim neighbor
		pint mask = pint(1).Dom((1 << i) - 1);
		r = mask.Select(r.Min(r.Rot(1 << i)), r);
	}
	return(r);
}

pfloat
pfloat::ScanMax() const
{
	// unary scan Mul (parallel prefix)
	pfloat r = *this;
	for (int i=0; i<REWAYS; ++i) {
		// send value up to i-dim neighbor
		pint mask = pint(1).Dom((1 << i) - 1);
		r = mask.Select(r.Max(r.Rot(1 << i)), r);
	}
	return(r);
}

// Gather

int
pfloat::Gather(float *p, const int n = REBITS) const
{
	// Gather a pfloat's entangled superposition
	// into the host array pointed to by p,
	// which is big enough for 2^ways elements
	// actual length of the array is returned
	int r = ((n > REBITS) ? REBITS : n);
	for (int e=0; e<r; ++e) {
		p[e] = Meas(e);
	}
	return(r);
}

// Display

void
pfloat::Show(const int vals = REBITS) const
{
	fprintf(stderr, "{");
	for (int i=0; i<(vals-1); ++i) {
		fprintf(stderr, "%g,%c", Meas(i), (((i%8)==7) ? '\n' : ' '));
	}
	fprintf(stderr, "%g}\n", Meas(vals-1));
}

