/*	pbpint.h

	Pattern INTeger layer for PBP.

	Note that pint can be either signed or unsigned, and of
	varying precision, with both attributes dynamically
	variable at runtime. This ensures that the minimum
	possible number of pbits are active in each computation.

	See pbp.h for copyright, etc.
*/

class pint
{
	uint8_t	type;		// 0 unsigned, 1 signed
	uint8_t	prec;		// precision in pbits
	pbit	bit[PINTBITS];	// actual pbits

public:
	// constructors
	pint() {
		// default to 1 unsigned pbit that's NaN
		type = UPINT;
		prec = 1;
		bit[0] = pbit(pbitNaN);
	}

	pint(const int konst) {
		// auto sizes to konst, also determines sign from konst
		type = (konst < 0);
		prec = PINTBITS;

		// install the bits
		forprec(i) {
			bit[i] = pbit(((1 << i) & konst) ? 1 : 0);
		}

		// get rid of any unnecessary bits
		*this = Minimize();
	}

	pint(const int konst, const int bits) {
		// make an integer with constant value konst
		// signed if bits is negative
		if (bits >= 0) {
			type = 0; // unsigned
			prec = bits;
		} else {
			type = 1; // signed
			prec = -bits;
		}

		// install bits
		forprec(i) {
			bit[i] = pbit(((1 << i) & konst) ? 1 : 0);
		}
	}

	operator pfloat() const; // forward decl

	// Logic
	pint LAnd(pint b) const;
	pint LOr(pint b) const;
	pint LXOr(pint b) const;

	// Bitwise
	pint And(pint b) const;
	pint Or(pint b) const;
	pint XOr(pint b) const;

	// Equality
	pint EQ(pint b) const;
	pint NE(pint b) const;

	// Relational
	pint GT(pint b) const;
	pint LT(pint b) const;
	pint GE(pint b) const;
	pint LE(pint b) const;

	// Shifts
	pint ShR(pint b) const;
	pint ShR(const int b) const;
	pint ShL(pint b) const;
	pint ShL(const int b) const;

	// Additive
	pint Add(pint b) const;
	pint Sub(pint b) const;

	// Multiplicative
	pint Mul(pint b) const;
	pint Mul(pint b, const int bits) const;
	pint Div(pint b) const;
	pint Rem(pint b) const; // C/C++ % is really remainder, not modulus

	// Not
	pint LNot() const;
	pint Not() const;
	pint Logic() const;	// like !!

	// Sign manipulation
	pint Neg() const;
	pint Abs() const;
	pint Signed() const;	// forces bits to be treated signed
	pint UnSigned() const;	// forces bits to be treated unsigned

	// PBP primitives
	pint Had(const int ways, const int dim) const;
	pint Rot(const int b) const;
	pint Flip(const int b) const;
	pint Reset(const int b) const;
	pint Set(const int e, const int v) const;
	pint Dom(const int b) const;
	int Meas(const int b) const;
	int Meas() const;
	REreg_t First() const;
	REreg_t Pop() const;

	// Helpers
	int Valid() const;
	pint Bit(const int b) const;
	pint Minimize() const;
	pint Extend(const int p) const;
	pint Promote(const pint& b) const;
	pint Select(pint vtrue, pint vfalse) const;
	pfloat Select(const pfloat& vtrue, const pfloat& vfalse) const;

	// Extended functions
	pint Min(pint b) const;
	pint Max(pint b) const;

	// Reductions
	int Any() const;
	int All() const;
	int ReduceAnd() const;
	int ReduceOr() const;
	int ReduceXOr() const;
	int ReduceAdd() const;
	int ReduceMul(const int preclimit) const; // problematic...
	int ReduceMin() const;
	int ReduceMax() const;

	// Scans (parallel prefix)
	pint ScanAnd() const;
	pint ScanOr() const;
	pint ScanXOr() const;
	pint ScanAdd() const;
	pint ScanMul(const int preclimit) const; // problematic...
	pint ScanMin() const;
	pint ScanMax() const;

	// Gather
	int Gather(int *p, const int n) const;

	// Display
	void Summary() const;
	void Show(const int vals) const;
	void ShowBits() const;

	// Friends
	friend class pbit;
	friend class pfloat;
	friend pint Cover(const int min, const int max, const int dim);
	friend pint Range(const int min, const int max, const int dim);
	friend pint Scatter(const int *p, const int n, const int reways);

	// Expensive library stuff
	pint SortUp() const;
	pint SortDown() const;

	// Assignment operators
	pint operator=(const pint& a) {
		type = a.type; prec = a.prec;
		forprec (i) bit[i] = a.bit[i];
		return(*this);
	}
	pint operator=(const int a) { return(*this = pint(a)); }
	pint operator*=(pint b) { *this = Mul(b); return(*this); }
	pint operator/=(pint b) { *this = Div(b); return(*this); }
	pint operator%=(pint b) { *this = Rem(b); return(*this); }
	pint operator+=(pint b) { *this = Add(b); return(*this); }
	pint operator-=(pint b) { *this = Sub(b); return(*this); }
	pint operator>>=(pint b) { *this = ShR(b); return(*this); }
	pint operator>>=(int b) { *this = ShR(b); return(*this); }
	pint operator<<=(pint b) { *this = ShL(b); return(*this); }
	pint operator<<=(int b) { *this = ShL(b); return(*this); }
	pint operator&=(pint b) { *this = And(b); return(*this); }
	pint operator|=(pint b) { *this = Or(b); return(*this); }
	pint operator^=(pint b) { *this = XOr(b); return(*this); }

	// Logical operators
	pint operator&&(const pint& b) const { return(LAnd(b)); }
	pint operator||(const pint& b) const { return(LOr(b)); }
	// there is no operator^^, but != effectively does it

	// Bitwise operators
	pint operator|(const pint& b) const { return(Or(b)); }
	pint operator^(const pint& b) const { return(XOr(b)); }
	pint operator&(const pint& b) const { return(And(b)); }

	// Equality operators
	pint operator==(const pint& b) const { return(EQ(b)); }
	pint operator!=(const pint& b) const { return(NE(b)); }

	// Relational operators
	pint operator>(const pint& b) const { return(GT(b)); }
	pint operator<(const pint& b) const { return(LT(b)); }
	pint operator>=(const pint& b) const { return(GE(b)); }
	pint operator<=(const pint& b) const { return(LE(b)); }

	// C++ 20 three-way comparison operator
	// uncomment if your C++ supports it (GCC doesn't yet)
//	pint operator<=>(const pint& b) const { return(GT(b) - LT(b)); }

	// Bitwise shift operators
	pint operator>>(pint b) const { return(ShR(b)); }
	pint operator>>(int b) const { return(ShR(b)); }
	pint operator<<(pint b) const { return(ShL(b)); }
	pint operator<<(int b) const { return(ShL(b)); }

	// Additive operators
	pint operator+(pint b) const { return(Add(b)); }
	pint operator-(pint b) const { return(Sub(b)); }

	// Multiplicative operators
	pint operator*(pint b) const { return(Mul(b)); }
	pint operator/(pint b) const { return(Div(b)); }
	pint operator%(pint b) const { return(Rem(b)); }

	// Not operators
	pint operator!() const { return(LNot()); }
	pint operator~() const { return(Not()); }

	// Unary plus and minus
	pint operator+() const { return(*this); }
	pint operator-() const { return(Neg()); }

	// Prefix increment and decrement
	pint& operator++() { *this = Add(pint(1)); return(*this); }
	pint& operator--() { *this = Sub(pint(1)); return(*this); }

	// Postfix increment and decrement
	// note that int arg is a C++ notation thing; there isn't an int arg
	pint operator++(int) { pint r = *this; *this = Add(pint(1)); return(r); }
	pint operator--(int) { pint r = *this; *this = Sub(pint(1)); return(r); }
};

// Logic

pint
pint::LAnd(pint b) const
{
	pint r = Logic();
	b = b.Logic();
	r.bit[0] = r.bit[0] & b.bit[0];
	return(r);
}

pint
pint::LOr(pint b) const
{
	pint r = Logic();
	b = b.Logic();
	r.bit[0] = r.bit[0] | b.bit[0];
	return(r);
}

pint
pint::LXOr(pint b) const
{
	pint r = Logic();
	b = b.Logic();
	r.bit[0] = r.bit[0] ^ b.bit[0];
	return(r);
}

// Bitwise

pint
pint::And(pint b) const
{
	pint r = Promote(b);
	b = b.Promote(r);
	for (int i=0; i<r.prec; ++i) {
		r.bit[i] = r.bit[i] & b.bit[i];
	}
	r = r.Minimize();
	return(r);
}

pint
pint::Or(pint b) const
{
	pint r = Promote(b);
	b = b.Promote(r);
	for (int i=0; i<r.prec; ++i) {
		r.bit[i] = r.bit[i] | b.bit[i];
	}
	r = r.Minimize();
	return(r);
}

pint
pint::XOr(pint b) const
{
	pint r = Promote(b);
	b = b.Promote(r);
	for (int i=0; i<r.prec; ++i) {
		r.bit[i] = r.bit[i] ^ b.bit[i];
	}
	return(r.Minimize());
}

// Equality

pint
pint::EQ(pint b) const
{
	pint r = XOr(b);
	r = r.Logic();
	r.bit[0] = ~r.bit[0];
	return(r);
}

pint
pint::NE(pint b) const
{
	pint r = XOr(b);
	r = r.Logic();
	return(r);
}

// Relational

pint
pint::GT(pint b) const
{
	pbit gt(0);
	pbit ep(1);
	pint r;
	r = Promote(b);
	b = b.Promote(r);

	if (r.type) {
		// signed > flips bits if they were negative
		r.bit[r.prec-1] = ~r.bit[r.prec-1];
		b.bit[r.prec-1] = ~b.bit[r.prec-1];
		r.type = UPINT;
		b.type = UPINT;
	}

	// unsigned > looks for msb that differs
	for (int i=r.prec-1; i>=0; --i) {
		gt = (gt | (ep & (r.bit[i] & ~b.bit[i])));
		ep = (ep & ~(r.bit[i] ^ b.bit[i]));
	}

	// always returns 1 bit unsigned
	r.bit[0] = gt;
	r.prec = 1;
	r.type = UPINT;
	return(r);
}

pint
pint::LT(pint b) const
{
	b = (b > *this);
	return(b);
}

pint
pint::GE(pint b) const
{
	b = ~(*this < b);
	return(b);
}

pint
pint::LE(pint b) const
{
	b = ~(*this > b);
	return(b);
}

// Shifts

pint
pint::ShR(pint b) const
{
	// shift right, treats shift amount as unsigned
	pint r = *this;

	for (int i=(b.prec-1); i>=0; --i) {
		for (int j=0; j<prec; ++j) {
			pbit x;

			if ((j + (1 << i)) >= prec) {
				x = (type ? bit[prec-1] : pbit(0));
			} else {
				x = r.bit[j + (1 << i)];
			}

			r.bit[j] = b.bit[i].Select(x, r.bit[j]);
		}
	}
	return(r.Minimize());
}

pint
pint::ShR(const int b) const
{
	// shift right, treats shift amount as unsigned
	pint r = *this;
	for (int i=0; i<r.prec; ++i) {
		r.bit[i] = (((i + b) < r.prec) ?
			    bit[i + b] :
			    (type ? bit[prec-1] : pbit(0)));
	}
	return(r.Minimize());
}

pint
pint::ShL(int b = 1) const
{
	// shift left, treats shift amount as unsigned
	int newprec = b + prec;
	if (newprec > REWAYS) newprec = REWAYS;
	pint r = Extend(newprec);
	for (int i=newprec-1; i>=0; --i) {
		r.bit[i] = (((i - b) >= 0) ? r.bit[i - b] : pbit(0));
	}
	return(r.Minimize());
}

pint
pint::ShL(const pint b) const
{
	// shift left, treats shift amount as unsigned
	pint r = *this;
	pint s(0);
	for (int i=0; i<b.prec; ++i) {
		s.bit[0] = b.bit[i];
		r = s.Select((r << (1<<i)), r);
	}
	return(r.Minimize());
}

// Additive

pint
pint::Add(pint b) const
{
	pint r = *this;
	r = r.Promote(b);
	r = r.Extend(r.prec + 1); // k bits + k bits = k+1 bits
	b = b.Promote(r);

	// ripple carry adder
	pbit c = pbit(0);
	for (int i=0; i<b.prec; ++i) {
		pbit x = r.bit[i] ^ b.bit[i];
		pbit n = r.bit[i] & b.bit[i];
		r.bit[i] = x ^ c;
		c = n | (x & c);
	}
	return(r.Minimize());
}

pint
pint::Sub(pint b) const
{
	pint r = *this;
	r = r.Promote(pint(-1)); // force promotion to signed
	r = r.Promote(b);
	r = r.Extend(r.prec + 1); // k bits + k bits = k+1 bits
	b = b.Promote(r);

	// ripple carry subtract
	pbit c = pbit(1);
	for (int i=0; i<b.prec; ++i) {
		pbit n = ~b.bit[i];
		pbit x = r.bit[i] ^ n;
		n = r.bit[i] & n;
		r.bit[i] = x ^ c;
		c = n | (x & c);
	}
	return(r.Minimize());
}

// Multiplicative

pint
pint::Mul(pint b) const
{
	// multiply
	pint v(0);
	pint a = *this;

	// deal with signed values
	if ((a.type == SPINT) || (b.type == SPINT)) {
		pint aneg = (a < pint(0));
		a = aneg.Select(-a, a);
		pint bneg = (b < pint(0));
		b = bneg.Select(-b, b);
		pint vneg = aneg ^ bneg;

		// faster if first arg has fewer bits
		for (int i=0; i<a.prec; ++i) {
			pint c(0,a.prec+b.prec);
			for (int j=0; ((j<b.prec)&&((i+j)<PINTBITS)); ++j) {
				c.bit[i+j] = b.bit[j] & a.bit[i];
			}
			c = c.Minimize();
			v = v + c;
		}

		v = vneg.Select(-v, v);
	} else {
		// faster if first arg has fewer bits
		for (int i=0; i<a.prec; ++i) {
			pint c(0,PINTBITS);
			for (int j=0; ((j<b.prec)&&((i+j)<PINTBITS)); ++j) {
				c.bit[i+j] = b.bit[j] & a.bit[i];
			}
			c = c.Minimize();
			v = v + c;
		}
	}
	return(v.Minimize());
}

pint
pint::Mul(pint b, const int bits) const
{
	// multiply, but only get least significant bits
	pint v(0);
	pint a = *this;

	// deal with signed values
	if ((a.type == SPINT) || (b.type == SPINT)) {
		// this is not really well defined...
		// signed values can change sign when limited
		pint aneg = (a < pint(0));
		a = aneg.Select(-a, a);
		pint bneg = (b < pint(0));
		b = bneg.Select(-b, b);
		pint vneg = aneg ^ bneg;
		if (b.prec > bits) b.prec = bits;
		int rounds = ((prec > bits) ? bits : prec);

		// faster if first arg has fewer bits
		for (int i=0; i<rounds; ++i) {
			pint c(0,a.prec+b.prec);
			for (int j=0; ((j<b.prec)&&((i+j)<PINTBITS)); ++j) {
				c.bit[i+j] = b.bit[j] & a.bit[i];
			}
			c = c.Minimize();
			v = v + c;
		}

		v = vneg.Select(-v, v);
	} else {
		if (b.prec > bits) b.prec = bits;
		int rounds = ((prec > bits) ? bits : prec);

		// faster if first arg has fewer bits
		for (int i=0; i<rounds; ++i) {
			pint c = b;

			for (int j=0; j<c.prec; ++j) {
				c.bit[j] = c.bit[j] & bit[i];
			}
			v = v + c;
			b = b << 1;

			// discard any extra bits
			if (v.prec > bits) v.prec = bits;
			if (b.prec > bits) b.prec = bits;
		}
	}

	return(v.Minimize());
}

pint
pint::Div(pint b) const
{
	// divide
	// divide by 0 yields 0 (maybe should be NaN?)
	pint s = (*this < pint(0));
	pint sb = (b < pint(0));
	pint rem = Abs();
	b = b.Abs();
	pint zero = (rem < b) | (b == pint(0));
	pint quo = 0;

	// shift & subtract loop (using adds)
	for (int i=rem.prec-1; i>=0; --i) {
		pint t = (b << i);
		pint le = (t <= rem);
		rem = rem - le.Select(t, pint(0));
		quo = quo | (le << i);
	}

	// force zeros, fix sign
	quo = zero.Select(pint(0), quo);
	quo = (s ^ sb).Select(-quo, quo);
	return(quo.Minimize());
}

pint
pint::Rem(pint b) const
{
	// C's % operator is called modulus,
	// but modulus should result in sign of divisor;
	// C's % is really remainder, with sign of dividend
	// modulus 0 yields 0 (maybe should be NaN?)
	pint s = (*this < pint(0));
	pint sb = (b < pint(0));
	pint rem = Abs();
	b = b.Abs();
	pint zero = (b == pint(0));

	// shift & subtract loop (using adds)
	for (int i=rem.prec-1; i>=0; --i) {
		pint t = (b << i);
		pint le = (t <= rem);
		rem = rem - le.Select(t, pint(0));
	}

	// force zeros, fix sign
	rem = zero.Select(pint(0), rem);
	rem = s.Select(-rem, rem);
	return(rem.Minimize());
}

// Not

pint
pint::LNot() const
{
	pint r = *this;
	r = r.Logic();
	r.bit[0] = ~r.bit[0];
	return(r);
}

pint
pint::Not() const
{
	pint r = *this;
	forprec(i) r.bit[i] = ~bit[i];
	return(r.Minimize());
}

pint
pint::Logic() const
{
	// reduce to a logical value, i.e., a single pbit
	pint r = *this;
	for (int i=1; i<prec; ++i) {
		r.bit[0] = r.bit[0] | r.bit[i];
	}
	r.type = 0; // always unsigned
	r.prec = 1; // single bit
	return(r);
}

// Sign manipulation

pint
pint::Neg() const
{
	pint r = pint(0).Sub(*this);
	return(r.Minimize());
}

pint
pint::Abs() const
{
	// absolute value (trivial if unsigned)
	if (type == UPINT) {
		return(*this);
	}

	// negate the negative ones
	pint t = (*this < 0);
	t = t.Select(-*this, *this);
	return(t.Minimize());
}

pint
pint::Signed() const
{
	// force signed; preserves bits, not value
	pint r = *this;
	r.type = SPINT;
	return(r);
}

pint
pint::UnSigned() const
{
	// force unsigned
	pint r = *this;
	r.type = UPINT;
	return(r);
}

// PBP primitives

pint
pint::Had(const int ways, const int dim = 0) const
{
	// make a bits-pbit Hadamard value using entanglement channels
	// starting with dimension dim and XOr it with this
	// negative ways forces signed
	int hbits = ((ways < 0) ? -ways : ways);
	int w = ((hbits > prec) ? hbits : prec); // extend this if needed
	pint r = Extend(w);
	int h = dim;

	// XOr the relevant (low) bits...
	for (int i=0; i<hbits; ++i) {
		r.bit[i] = r.bit[i] ^ pbit(pbitH(h));
		++h;
	}
	if (ways < 0) r = r.Signed();
	return(r.Minimize());
}

pint
pint::Rot(const int b) const
{
	pint r = *this;
	forprec(i) r.bit[i] = bit[i].Rot(b);
	return(r);
}

pint
pint::Flip(const int b) const
{
	pint r = *this;
	forprec(i) r.bit[i] = bit[i].Flip(b);
	return(r);
}

pint
pint::Reset(const int b) const
{
	pint r = *this;
	forprec(i) r.bit[i] = bit[i].Reset(b);
	r = r.Minimize();
	return(r);
}

pint
pint::Set(const int e, const int v = 1) const
{
	// make a pint of v so that we know type and prec
	// to promote this to before set/reset bits
	pint t(v);
	pint r = Promote(t); // make r big enough to hold v
	for (int i=0; i<r.prec; ++i) {
		if (v & (1 << i)) {
			r.bit[i] = r.bit[i].Set(e);
		} else {
			r.bit[i] = r.bit[i].Reset(e);
		}
	}
	return(r.Minimize());
}

pint
pint::Dom(const int b) const
{
	pint r = *this;
	forprec(i) r.bit[i] = bit[i].Dom(b);
	return(r);
}

int
pint::Meas(const int b) const
{
	// note that Measurement always produces 0 or 1
	// bit values, never superpositions, which is why
	// this can return an ordinary int value
	register int v = 0;
	forprec(i) if ((bit[i].Meas(b)).v) v |= (1<<i);
	if (type == SPINT) {
		// perform sign extension?
		if (v & (1<<(prec-1))) {
			// yes, negative value
			v |= ~((1<<(prec-1))-1);
		}
	}
	return(v);
}

int
pint::Meas() const
{
	// Quantum-like random measurement
	return(Meas(rand() & (REBITS - 1)));
}

REreg_t
pint::First() const
{
	// first is first 1 in any pbit
	register REreg_t first = REBITS;
	forprec(i) {
		register REreg_t t = bit[i].First();
		if (t < first) first = t;
	}
	return(first);
}

REreg_t
pint::Pop() const
{
	// pop(ulation) is number of non-zero multi-bit values
	pint r = *this;
	r = r.Logic();
	return(r.bit[0].Ones());
}

// Helpers

int
pint::Valid() const {
	// check to see if this pbit is valid
	if (type > 1) return(0);
	if ((prec < 1) || (prec >= PINTBITS)) return(0);
	forprec (i) {
		if (!bit[i].Valid()) return(0);
	}
	return(1);
}

pint
pint::Bit(const int b) const
{
	// Extract a particular bit position
	pint r(1);
	r.bit[0] = ((b < prec) ? bit[b] : pbit(0));
	return(r);
}

pint
pint::Minimize() const
{
	// make value as small as possible
	pint r = *this;

	if (r.type) {
		// signed?
		if ((r.prec < 2) || (r.bit[r.prec-1].v != 0)) {
			// still signed
			for (int i=r.prec-1; i>0; --i) {
				if (r.bit[i].v != r.bit[i-1].v) return(r);
				--r.prec;
			}
			return(r);
		}

		// make unsigned
		r.type = UPINT;
		--(r.prec);
	}

	// unsigned
	for (int i=r.prec-1; i>0; --i) {
		if (r.bit[i].v) return(r);
		--r.prec;
	}
	return(r);
}

pint
pint::Extend(const int p) const
{
	// Extend to precision p, truncate if already bigger
	int pnew = ((p > PINTBITS) ? PINTBITS : p);
	pint r = *this;

	// Signed/unsigned extend
	if (r.prec < pnew) {
		for (int i=r.prec; i<pnew; ++i) {
			r.bit[i].v = (r.type ? r.bit[i-1].v : 0);
		}
	}
	r.prec = pnew;
	return(r);
}

pint
pint::Promote(const pint& b) const
{
	// promote this to covering type for this, b
	register int p = ((prec < b.prec) ? b.prec : prec);
	pint r = *this;

	// if the signs differ and the signed one isn't bigger
	// than the unsigned one, add a bit and make both signed
	if (((type && (!b.type)) && (prec <= b.prec)) ||
	    ((b.type && (!type)) && (b.prec <= prec))) {
		++p;
	}

	// Extend precision
	r = r.Extend(p);

	// Correct sign
	r.type = (type | b.type);
	return(r);
}

pint
pint::Select(pint vtrue, pint vfalse) const
{
	// multiplexor
	pint r = Logic(); // make 1-bit logical value
	pbit c = r.bit[0];
	vtrue = vtrue.Promote(vfalse);
	vfalse = vfalse.Promote(vtrue);
	r = vtrue;
	for (int i=0; i<vtrue.prec; ++i) {
		r.bit[i] = c.Select(vtrue.bit[i], vfalse.bit[i]);
	}
	return(r.Minimize());
}

// Extended functions

pint
pint::Min(pint b) const
{
	pint r = Promote(b);
	b = b.Promote(r);
	pint gt = (r > b);
	for (int i=0; i<b.prec; ++i) {
		r.bit[i] = gt.bit[0].Select(b.bit[i], r.bit[i]);
	}
	return(r.Minimize());
}

pint
pint::Max(pint b) const
{
	pint r = Promote(b);
	b = b.Promote(r);
	pint gt = (r > b);
	for (int i=0; i<b.prec; ++i) {
		r.bit[i] = gt.bit[0].Select(r.bit[i], b.bit[i]);
	}
	return(r.Minimize());
}

// Reductions

int
pint::Any() const
{
	// this is the classic SIMD test
	// true if any entanglement channel value is non-0
	pint r = *this;
	r = r.Logic();
	return(r.bit[0].v ? 1 : 0);
}

int
pint::All() const
{
	// this is the classic SIMD test
	// true is all entanglement channels values are non-0
	pint r = *this;
	r = r.Logic();
	return((r.bit[0].v == 1) ? 1 : 0);
}

int
pint::ReduceAnd() const
{
	// unary bitwise AND reduction
	int r = 0;
	forprec(i) r |= ((bit[i].v == pbit1) << i);
	if ((type == SPINT) && (r & (1 << (prec-1)))) {
		// sign extend to make it negative value
		r |= ~((1 << (prec-1)) - 1);
	}
	return(r);
}

int
pint::ReduceOr() const
{
	// unary bitwise OR reduction
	int r = 0;
	forprec(i) r |= ((bit[i].v != pbit0) << i);
	if ((type == SPINT) && (r & (1 << (prec-1)))) {
		// sign extend to make it negative value
		r |= ~((1 << (prec-1)) - 1);
	}
	return(r);
}

int
pint::ReduceXOr() const
{
	// unary bitwise OR reduction
	int r = 0;
	forprec(i) r |= ((bit[i].Ones() & 1) << i);
	if ((type == SPINT) && (r & (1 << (prec-1)))) {
		// sign extend to make it negative value
		r |= ~((1 << (prec-1)) - 1);
	}
	return(r);
}

int
pint::ReduceAdd() const
{
	// unary add reduction (sum)
	int r = 0;
	if (type == UPINT) {
		// unsigned does weighted add
		for (int i=prec-1; i>=0; --i) {
			// sums weighted population counts
			r += r + bit[i].Ones();
		}
	} else {
		// signed does pos & neg sums separately
		pint uns = (*this > pint(0));
		pint pos = uns.Select(*this, pint(0));
		pint neg = -uns.Select(pint(0), *this);
		for (int i=prec-1; i>=0; --i) {
			// sums weighted population counts
			r += r + pos.bit[i].Ones() - neg.bit[i].Ones();
		}
	}
	return(r);
}

int
pint::ReduceMul(const int preclimit = REWAYS) const
{
	// Reduce Mul by classical tree reduction
	// does large steps first because that makes
	// more of the chunks 0 early
	// Be careful! It's easy to go out of pint range
	pint r = *this;
	for (int i=REWAYS-1; i>0; i>>=1) {
		pint tomul = r.Rot(1<<i);
		pint mask = pint(0).Dom(1<<i);
		r = r.Mul(mask.Select(tomul, pint(0)), preclimit);
	}
	return(r.Meas(0));
}

int
pint::ReduceMin() const
{
	// unary minimum reduction
	pint t = *this;
	int r = ((1 << prec) - 1);

	if (type == SPINT) {
		// make it unsigned by adding max positive + 1
		t.bit[prec-1] = t.bit[prec-1] ^ pbit(1);
		t.type = UPINT;
	}

	// from MSB down
	for (int i=prec-1; i>=0; --i) {
		// anybody got a 0 in this position?
		if (t.bit[i].v != pbit1) {
			// set lower bits that don't match uppers
			for (int j=0; j<i; ++j) {
				t.bit[j] = t.bit[j] | t.bit[i];
			}
			r ^= (1 << i);
		}
	}

	if (type == SPINT) {
		// if it was signed, subtract max positive
		r -= (1 << (prec - 1));
	}
	return(r);
}

int
pint::ReduceMax() const
{
	// unary maximum reduction
	pint t = *this;
	int r = 0;

	if (type == SPINT) {
		// make it unsigned by adding max positive + 1
		t.bit[prec-1] = t.bit[prec-1] ^ pbit(1);
		t.type = UPINT;
	}

	// from MSB down
	for (int i=prec-1; i>=0; --i) {
		// anybody got a 1 in this position?
		if (t.bit[i].v != pbit0) {
			// wipe lower bits that don't match uppers
			for (int j=0; j<i; ++j) {
				t.bit[j] = t.bit[j] & t.bit[i];
			}
			r |= (1 << i);
		}
	}

	if (type == SPINT) {
		// if it was signed, subtract max positive
		r -= (1 << (prec - 1));
	}
	return(r);
}

// Scans (parallel prefix)

pint
pint::ScanAnd() const
{
	// unary bitwise AND scan (parallel prefix)
	pint r = *this;
	forprec(i) {
		int pos = (~bit[i]).First();
		switch (pos) {
		case 0: r.bit[i] = pbit(0); break;
		case REBITS: break; // already pbit(1)
		default:
			pbit t(0);
			t = t.Dom(pos-1);
			r.bit[i] = t;
		}
	}
	return(r);
}

pint
pint::ScanOr() const
{
	// unary bitwise OR scan (parallel prefix)
	pint r = *this;
	forprec(i) {
		int pos = bit[i].First();
		switch (pos) {
		case 0: r.bit[i] = pbit(1); break;
		case REBITS: break; // already pbit(0)
		default:
			pbit t(1);
			t = t.Dom(pos-1);
			r.bit[i] = t;
		}
	}
	return(r);
}

pint
pint::ScanXOr() const
{
	// unary scan XOr (parallel prefix)
	pint r = *this;
	for (int i=0; i<REWAYS; ++i) {
		// send value up to i-dim neighbor
		pint mask = (pint(1)).Dom((1 << i) - 1);
		r = mask.Select(r ^ r.Rot(1 << i), r);
	}
	return(r);
}

pint
pint::ScanAdd() const
{
	// unary scan Add (parallel prefix)
	pint r = *this;
	for (int i=0; i<REWAYS; ++i) {
		// send value up to i-dim neighbor
		pint mask = pint(1).Dom((1 << i) - 1);
		r = mask.Select(r + r.Rot(1 << i), r);
	}
	return(r);
}

pint
pint::ScanMul(const int preclimit = REWAYS) const
{
	// unary scan Mul (parallel prefix)
	// note that this can easily exceed the range of
	// representable integers... that's not a bug
	pint r = *this;
	for (int i=0; i<prec; ++i) {
		// send value up to i-dim neighbor
		pint mask = (pint(1)).Dom((1 << i) - 1);
		r = mask.Select(r.Mul(r.Rot(1 << i), preclimit), r);
	}
	return(r);
}

pint
pint::ScanMin() const
{
	// unary scan Max (parallel prefix)
	pint r = *this;
	for (int i=0; i<REWAYS; ++i) {
		// send value up to i-dim neighbor
		pint mask = (pint(1)).Dom((1 << i) - 1);
		r = r.Min(mask.Select(r.Rot(1 << i), r));
	}
	return(r);
}

pint
pint::ScanMax() const
{
	// unary scan Max (parallel prefix)
	pint r = *this;
	for (int i=0; i<REWAYS; ++i) {
		// send value up to i-dim neighbor
		pint mask = (pint(1)).Dom((1 << i) - 1);
		r = r.Max(mask.Select(r.Rot(1 << i), r));
	}
	return(r);
}

// Gather

int
pint::Gather(int *p, const int n) const
{
	// Gather a pint's entangled superposition
	// into the host array pointed to by p,
	// which is big enough for 2^ways elements
	// actual length of the array is returned
	int r = ((n > REBITS) ? REBITS : n);
	for (int e=0; e<r; ++e) {
		p[e] = Meas(e);
	}
	return(r);
}

// Display

void
pint::Summary() const
{
	// summarize this pint
	fprintf(stderr, "%s%d_t {", (type ? "int" : "uint"), prec);
	forprec (i) {
		fprintf(stderr, "%d%s", bit[i].v, ((i==(prec-1)) ? "}\n" : ","));
	}
}

void
pint::Show(const int vals = REBITS) const
{
	// show bits in this pint
	Summary();
	fprintf(stderr, "{");
	for (int i=0; i<vals-1; ++i) {
		fprintf(stderr, "%d,%c", Meas(i), (((i%16)==15) ? '\n' : ' '));
	}
	fprintf(stderr, "%d}\n", Meas(vals-1));
}

void
pint::ShowBits() const
{
	// show bits in this pint
	Summary();
	for (int i=0; i<prec; ++i) bit[i].Show();
}

// Friends

pint
Cover(const int min, const int max, const int dim = 0)
{
	// create an entangled superposition that covers [min..max]
	int range = (max - min) + 1;
	int bits = 0;
	while ((1 << (bits-1)) <= range) ++bits;
	++bits;
	return(pint(0).Had(bits, dim) + pint(min));
}

pint
Range(const int min, const int max, const int dim = 0)
{
	// create an entangled superposition that TIGHTLY covers [min..max]
	// excess values at the end are made into all 0
	pint r = Cover(min, max, dim);
	pint mask(0);
	mask = mask.Dom(max - min);
	return(mask.Select(r, pint(0)));
}

pint
Scatter(const int *p, const int n, const int reways = REWAYS)
{
	// Scatter a host array into the
	// entanglement channels of a pint
	int sawneg = 0; // any negative values?
	for (int j=0; j<n; ++j) {
		if (p[j] < 0) {
			sawneg = 1;
			break;
		}
	}
	pint r(0, reways);
	int i = 0;
	do {
		// what should first bit be?
		int prev = (p[0] & (1 << i));
		int t = (prev != 0);
		int state = 0;

		// keep flipping all below when we
		// see a change (domino)
		for (int j=1; j<n; ++j) {
			int now = (p[j] & (1 << i));
			if (now != prev) {
				r.bit[i] = r.bit[i].Dom(j-1);
				state = !state;
				prev = now;;
			}
		}

		// did we end right?
		if (state != t) {
			// one more flip of everything
			r.bit[i] = r.bit[i].Dom(n-1);
		}
	} while (++i < reways);

	if (sawneg) r = r.Signed();
	return(r.Minimize());
}

// Expensive library stuff

pint
pint::SortUp() const
{
	// use Flip to implement Bitonic sort...
	// of values in entanglement channels
	// this does NOT change probabilities
	// sorts ascending unless non-0 arg given
	pint r = *this;
	pint i;
	i = pint(0).Had(prec, 0);

	for (int k=2; k<=REBITS; k+=k) {
		for (int j=k>>1; j>0; j=j>>1) {
			pint mate = r.Flip(j);
			pint dir = (((i ^ pint(j)) > i) ^
				    ((i & pint(k & (REBITS-1))).Logic()));
			pint swap = ((r < mate) != dir) && ((r > mate) == dir);
			r = swap.Select(mate, r);
		}
	}
	return(r);
}

pint
pint::SortDown() const
{
	// use Flip to implement Bitonic sort...
	// of values in entanglement channels
	// this does NOT change probabilities
	// sorts ascending unless non-0 arg given
	pint r = *this;
	pint i;
	i = pint(0).Had(prec, 0);

	for (int k=2; k<=REBITS; k+=k) {
		for (int j=k>>1; j>0; j=j>>1) {
			pint mate = r.Flip(j);
			pint dir = (((i ^ pint(j)) > i) ^
				    ((i & pint(k & (REBITS-1))).Logic()));
			pint swap = ((r < mate) == dir) && ((r > mate) != dir);
			r = swap.Select(mate, r);
		}
	}
	return(r);
}
