/*	pbpre.h

	Regular Expression layer for PBP.

	See pbp.h for copyright, etc.

	Overview: The RE layer basically contains the complete
	structure for everything below the pbit layer. The RE
	registers are allocated as a stack. Each potentially new
	entry is made on the top of the stack, but uniq() tests
	if there is a copy already, and doesn't keep the new top
	of stack one if there is.
*/

class RE
{
	AoB		aob; // AoB and applicative cache
	AoBreg_t	reg[REREGS][REAOBS];
	REreg_t		index[AOBREGS]; // 0 means empty, >0 means reg[index-1]
	REreg_t		regs;
	int		hit, rehash;

public:
	// Helpers
	void copy(const REreg_t d, const REreg_t s) {
		memcpy(&(reg[d][0]), &(reg[s][0]), (REAOBS * sizeof(AoBreg_t)));
	}

	int match(const REreg_t a) {
		forREAOBS(i) {
			if (reg[a][i] != reg[regs][i]) return(0);
		}
		return(1);
	}

	void Stats() {
		aob.Stats();
		fprintf(stderr, "RE: %d regs, %d hits, %d rehashes\n", regs, hit, rehash);
	}

	void Show(const REreg_t a) {
		forREAOBS(i) {
			aob.Show(reg[a][i]);
		}
	}

	// Constructor
	RE() {
		// initialization patterns for words
		AoBword_t	inits[] = { 0x0000000000000000LL,
				    0xffffffffffffffffLL,
				    0xaaaaaaaaaaaaaaaaLL,
				    0xccccccccccccccccLL,
				    0xf0f0f0f0f0f0f0f0LL,
				    0xff00ff00ff00ff00LL,
				    0xffff0000ffff0000LL,
				    0xffffffff00000000LL };

		// set-up Stats
		hit = 0;
		rehash = 0;

		// set-up 0, 1, H(0), H(1), H(2), ...
		bzero(&(index[0]), (REREGS * sizeof(index[0])));
		regs = 0;
		while (regs < (REWAYS+2)) {
			forREAOBS(i) {
				reg[regs][i] = ((regs < (AOBWAYS+2)) ? regs : ((i & (1<<(regs-(AOBWAYS+2)))) != 0));
			}

			uniq(); // making unique entry also increments regs
		}
	}

	// Uniqueness functions
	REreg_t hash() {
		// should really pick this by benchmarking...
		REreg_t h = 42; // a nice random starting bit pattern

		for (int i=0; i<REAOBS; ++i) {
			h = (h << 3) | (h & 7);
			h ^= reg[regs][i];
		}

		return((h ^ (h / REREGS)) & (REREGS-1));
	}

	REreg_t uniq() {
		REreg_t h = hash();
		for (;;) {
			// empty entry
			if (!index[h]) { index[h] = ++regs; return(regs - 1); }

			// found it
			if (index[h] && match(index[h] - 1)) { ++hit; return(index[h] - 1); }

			++rehash;
			h = ((h + 347) & (REREGS-1));
		}
	}

	// Basic operations
	REreg_t And(const REreg_t a, const REreg_t b) {
		// normalize
		register REreg_t ta, tb;
		if (a <= b) { ta = a; tb = b; } else { ta = b; tb = a; }

		// check special cases
		if (!ta) return(0);
		if ((ta == 1) || (ta == tb)) return(tb);

		// do AoB operation
		forREAOBS(i) reg[regs][i] = aob.And(reg[ta][i], reg[tb][i]);
		return(uniq());
	}

	REreg_t Or(const REreg_t a, const REreg_t b) {
		// normalize
		register REreg_t ta, tb;
		if (a <= b) { ta = a; tb = b; } else { ta = b; tb = a; }

		// check special cases
		if ((!ta) || (ta == tb)) return(tb);
		if (ta == 1) return(1);

		// do AoB operation
		forREAOBS(i) reg[regs][i] = aob.Or(reg[ta][i],  reg[tb][i]);
		return(uniq());
	}

	REreg_t Xor(const REreg_t a, const REreg_t b) {
		// normalize
		register REreg_t ta, tb;
		if (a <= b) { ta = a; tb = b; } else { ta = b; tb = a; }

		// check special cases
		if (!ta) return(tb);
		if (ta == tb) return(0);

		// do AoB operation
		forREAOBS(i) reg[regs][i] = aob.Xor(reg[ta][i], reg[tb][i]);
		return(uniq());
	}

	REreg_t Rot(const REreg_t a, const REreg_t b) {
		// normalize
		register REreg_t tb = (b & (REBITS-1));

		// check special case
		if (!tb) return(a);

		// first, shift by AoBs
		int ws = (tb / AOBBITS);
		forREAOBS(i) reg[regs][i] = reg[a][(i - ws) & (REAOBS-1)];

		// then by bits if any left
		int br = (tb % AOBBITS);
		if (br) {
			// rotate 'em individually, which makes top bits correct
			forREAOBS(i) reg[regs+1][i] = aob.Rot(reg[regs][i], br);

			// copy bits from other AoBs
			AoBreg_t lomask = aob.Dom(0, br-1);
			AoBreg_t himask = aob.Xor(1, lomask);
			forREAOBS(i) {
				reg[regs][i] = aob.Or(aob.And(himask, reg[regs+1][i]),
						      aob.And(lomask, reg[regs+1][(i - 1) & (REAOBS - 1)]));
			}
		}

		// return unique
		return(uniq());
	}

	REreg_t Flip(const REreg_t a, const REreg_t b) {
		// normalize
		register REreg_t tb = (b & (REBITS-1));

		// check special case
		if (!tb) return(a);

		// first, copy into work area
		copy(regs, a);

		// do any within-an-AoB steps
		forREAOBS(i) {
			reg[regs][i] = aob.Flip(reg[a][i], (tb & (AOBBITS-1)));
		}

		// now do across AoBs
		for (int k=0; k<(REWAYS-AOBWAYS); ++k) {
			if (((1 << AOBWAYS) << k) & tb) {
				// copy into regs+1
				copy(regs+1, regs);

				// re-arrange into regs
				forAOBWORDS(i) {
					reg[regs][i] = reg[regs+1][i ^ (1 << k)];
				}
			}
		}

		// return unique
		return(uniq());
	}

	REreg_t Reset(const REreg_t a, const REreg_t b) {
		// normalize
		register REreg_t tb = (b & (REBITS-1));

		// do bitwise operation
		copy(regs, a);
		reg[regs][tb >> AOBWAYS] = aob.Reset(reg[regs][tb >> AOBWAYS], (tb & (AOBBITS-1)));
		return(uniq());
	}

	REreg_t Set(const REreg_t a, const REreg_t b) {
		// normalize
		register REreg_t tb = (b & (REBITS-1));

		// do bitwise operation
		copy(regs, a);
		reg[regs][tb >> AOBWAYS] = aob.Set(reg[regs][tb >> AOBWAYS], (tb & (AOBBITS-1)));
		return(uniq());
	}

	REreg_t Tog(const REreg_t a, const REreg_t b) {
		// normalize
		register REreg_t tb = (b & (REBITS-1));

		// do bitwise operation
		copy(regs, a);
		reg[regs][tb >> AOBWAYS] = aob.Tog(reg[regs][tb >> AOBWAYS], (tb & (AOBBITS-1)));
		return(uniq());
	}

	REreg_t Dom(const REreg_t a, const REreg_t b) {
		// normalize
		register REreg_t tb = (b & (REBITS-1));

		// do bitwise operation
		copy(regs, a);
		for (int i=0; i<(tb >> AOBWAYS); ++i) reg[regs][i] = aob.Xor(1, reg[regs][i]);
		reg[regs][tb >> AOBWAYS] = aob.Dom(reg[regs][tb >> AOBWAYS], (tb & (AOBBITS-1)));
		return(uniq());
	}

	REreg_t Meas(const REreg_t a, const REreg_t b) {
		// normalize
		register REreg_t tb = (b & (REBITS-1));
		// result is 0 or 1; no need to do uniq()
		return(aob.Meas(reg[a][tb >> AOBWAYS], (tb & (AOBBITS-1))));
	}

	REreg_t Any(const REreg_t a) {
		// result is 0 or 1; no need to do uniq()
		forREAOBS(i) {
			if (aob.Any(reg[a][i])) return(1);
		}
		return(0);
	}

	REreg_t All(const REreg_t a) {
		// result is 0 or 1; no need to do uniq()
		forREAOBS(i) {
			if (!aob.All(reg[a][i])) return(0);
		}
		return(1);
	}

	REreg_t First(const REreg_t a) {
		// check special cases... initialized constants
		if (a < (REWAYS+2)) {
			if (a == 0) return(1 << REWAYS);
			if (a == 1) return(0);
			return(1 << (a - 2));
		}

		// result is an integer
		register REreg_t r = 0;
		forREAOBS(i) {
			register REreg_t t = aob.First(reg[a][i]);
			r += t;
			if (t < AOBBITS) {
				return(r);
			}
		}
		return(r);
	}

	REreg_t Ones(const REreg_t a) {
		// check special cases... initialized constants
		if (a < (REWAYS+2)) {
			if (a == 0) return(0);
			if (a == 1) return(1 << REWAYS);
			return(1 << (REWAYS - 1));
		}
		
		// result is an integer
		REreg_t pop = 0;
		forREAOBS(i) {
			pop += aob.Ones(reg[a][i]);
		}
		return(pop);
	}
};

RE re;	// instantiate everything
